// InstDriverXP.c
//;-------------------------------------------------------------------------
//; Copyright (c) 2003 Steven Freyder, David Helland, and Bruce D. Lightner.
//; Use subject to licensing restrictions described in "PICOWEB SERVER
//; SOFTWARE LICENSE AGREEMENT DATED JULY 14, 1999" (See file LICENSE.tt).
//;-------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <winioctl.h>
#include <process.h>

#include "SetPrivIoXP.h"
#include "InstDriverXP.h"

#include "DriverSys.h"       /* driver's .SYS file binary */

#define PRINTF PrintfPrivIoXP

BOOL StartDriverXP(
    BOOL bVerbose,
    char *szPrefix,
    char *szDriverName
){
    int tries;
    SC_HANDLE SchSCManager = NULL;
    SC_HANDLE schService = NULL;

    // Open Handle to Service Control Manager
    if (!(SchSCManager = OpenSCManager(NULL,	// machine (NULL == local)
				 NULL,	// database (NULL == default)
				 SC_MANAGER_ALL_ACCESS)	// access required
        )) {
	ShowErr("OpenSCManager");
	if (GetLastError() == ERROR_ACCESS_DENIED) {
	    PRINTF("%sno access rights Service Control Manager!\n", szPrefix);
	} else {
	    PRINTF("%serror opening Service Control Manager!\n", szPrefix);
	}
	PRINTF("%s%s driver not installed or started!\n", szPrefix, szDriverName);
	PRINTF(
            "%sHint: Ask administrator to install %s driver \n", szPrefix, szDriverName);
	goto abort;
    }
    for (tries = 1; tries <= 2; ++tries) {
	// Open a Handle to the Service Database
	if (!(schService = OpenService(SchSCManager, szDriverName, SERVICE_ALL_ACCESS))) {
	    switch (GetLastError()) {
	    case ERROR_ACCESS_DENIED:
                ShowErr("OpenService");
		PRINTF("%sneed access rights to %s service database!\n", szPrefix, szDriverName);
	        PRINTF("%sHint: Ask administrator to install %s driver \n", szPrefix, szDriverName);
		goto abort;
	    case ERROR_INVALID_NAME:
                ShowErr("OpenService");
		PRINTF("%sservice name \"%s\" invalid!\n", szPrefix, szDriverName);
		goto abort;
	    case ERROR_SERVICE_DOES_NOT_EXIST:
		PRINTF("%s%s driver not installed!\n", szPrefix, szDriverName);
                if (tries == 1) {
		    PRINTF("%sInstalling %s driver\n", szPrefix, szDriverName);
		    if (!InstallDriverXP(bVerbose, szPrefix, szDriverName, 
                                  DriverSys, sizeof(DriverSys))) goto abort;
                } else {
		    goto abort;
                }
		break;
	    default:
                ShowErr("OpenService");
		PRINTF("%sOpenService failed!\n", szPrefix);
		goto abort;
	    }
        } else {
            break;
        }
    }

    if (StartService(schService, 0, NULL)) {
        if (bVerbose) {
	    PRINTF("%s%s driver started (0x%lx)\n", szPrefix,
                            szDriverName, (DWORD)schService);
        }
    } else {
	if (GetLastError() == ERROR_SERVICE_ALREADY_RUNNING) {
	    PRINTF("%s%s driver already running\n", szPrefix, szDriverName);
            goto finis;
	} else {
            ShowErr("StartService");
	    PRINTF("%serror starting %s driver service!\n", szPrefix, szDriverName);
	    goto abort;
	}
    }

finis:
    if (schService) CloseServiceHandle(schService);
    return (TRUE);

abort:
    if (schService) CloseServiceHandle(schService);
    return (FALSE);
}

BOOL InstallDriverXP(
    BOOL bVerbose,
    char *szPrefix,
    char *szDriverName,
    unsigned char *uCode,
    int nCodeBytes
){
    SC_HANDLE SchSCManager = NULL;
    SC_HANDLE schService = NULL;
    HANDLE hFile = INVALID_HANDLE_VALUE; 
    DWORD dwNumBytesWritten;
    char szDriverFileName[MAX_PATH+32];

    if (!MakeFullDriverFileName(szDriverName,
                                szDriverFileName, sizeof(szDriverFileName))) {
        ShowErr("MakeFullDriverFileName");
	PRINTF("%serror making driver file name!\n", szPrefix);
        goto abort;
    }

#ifndef DO_NOT_INSTALL_SYS_FILE
    if (!(hFile = CreateFile(szDriverFileName,
                     GENERIC_WRITE,
                     0,
                     (LPSECURITY_ATTRIBUTES)NULL,
                     CREATE_ALWAYS,
                     FILE_ATTRIBUTE_NORMAL,
                     NULL))) {
        ShowErr("szDriverFileName");
	PRINTF("%sFailed to copy driver to %s!\n", szPrefix, szDriverFileName);
        goto abort;
    }
    if (!WriteFile(hFile, uCode, nCodeBytes, &dwNumBytesWritten, 0)) {
        ShowErr("szDriverFileName");
	PRINTF("%sFailed to copy driver to %s!\n", szPrefix, szDriverFileName);
        goto abort;
    }
    if (!CloseHandle(hFile)) {
        ShowErr("szDriverFileName");
	PRINTF("%sFailed to close file %s!\n", szPrefix, szDriverFileName);
        goto abort;
    }
    PRINTF("%s%s driver installed at %s\n", szPrefix,
                        szDriverName, szDriverFileName);
#endif /* !DO_NOT_INSTALL_SYS_FILE */

    SchSCManager = OpenSCManager(NULL,	// machine (NULL == local)
				 NULL,	// database (NULL == default)
				 SC_MANAGER_ALL_ACCESS);	// access required

    // Create service/driver - this process adds registry keys under
    // HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services
    // Note: No checks if driver .SYS file exists, or if path is correct!
    if (!(schService = CreateService(SchSCManager,	// SCManager database
			       szDriverName,	// name of service
			       szDriverName,	// name to display
			       SERVICE_ALL_ACCESS,	// desired access
			       SERVICE_KERNEL_DRIVER,	// service type
#if 1
			       SERVICE_DEMAND_START,	// start type
#else
			       SERVICE_AUTO_START,	// start type
#endif
			       SERVICE_ERROR_NORMAL,	// error control type
			       szDriverFileName,	// service's binary
			       NULL,	// no load ordering group
			       NULL,	// no tag identifier
			       NULL,	// no dependencies
			       NULL,	// LocalSystem account
			       NULL)	// no password
	                       )) {
	if (GetLastError() == ERROR_SERVICE_EXISTS) {
	    PRINTF("%s%s driver already exists! No action taken.\n", szPrefix, szDriverName);
            goto finis;
	} else {
            ShowErr("CreateService");
	    PRINTF("%serror creating service!\n", szPrefix);
            goto abort;
        }
    } else {
        if (bVerbose) {
	    PRINTF("%s%s service installed using %s\n", szPrefix,
                            szDriverName, szDriverFileName);
        }
    }

finis:
    if (schService) CloseServiceHandle(schService);
    return(TRUE);

abort:
    if (schService) CloseServiceHandle(schService);
    return(FALSE);
}


BOOL RemoveDriverXP(
    BOOL bVerbose,
    char *szPrefix,
    char *szDriverName,
    BOOL bDelDriverFile
){
    BOOL nExitStatus = FALSE;  // default to error exit
    SC_HANDLE hSCManager = NULL;
    SC_HANDLE hService = NULL;
    SERVICE_STATUS serviceStatus;
    HANDLE hFile;
    char szDriverFileName[MAX_PATH+32];

    if (bVerbose) PRINTF("%suninstalling Windows %s driver service\n", szPrefix, szDriverName);

    if (!(hSCManager = OpenSCManager(
			NULL,	// machine (local)
			NULL,	// database (SERVICES_ACTIVE_DATABASE)
			SC_MANAGER_ALL_ACCESS)	// type of access
	)) {
	ShowErr("OpenSCManager");
	PRINTF("Error opening Service Control Manager!\n");
	goto abort;
    }

    if (!(hService = OpenService(hSCManager, szDriverName, SERVICE_ALL_ACCESS))) {
	ShowErr("OpenService");
	PRINTF("%sError opening %s driver service!\n", szPrefix, szDriverName);
	PRINTF("%sHint: Was %s driver really installed?\n", szPrefix, szDriverName);
	goto abort;
    }

#if 0
    if (!QueryServiceStatus(hService, &serviceStatus)) {
	ShowErr("QueryServiceStatus");
	goto abort;
    }
    PRINTF("%sdwCurrentState=0x%x\n", szPrefix, serviceStatus.dwCurrentState);
#endif

    if (ControlService(hService, SERVICE_CONTROL_STOP, &serviceStatus)) {
	if (bVerbose) PRINTF("%s%s driver service stopped\n", szPrefix, szDriverName);
    } else if (GetLastError() == ERROR_SERVICE_NOT_ACTIVE) {
	if (bVerbose) PRINTF("%s%s driver service already stopped\n", szPrefix, szDriverName);
    } else {
	ShowErr("ControlService(SERVICE_CONTROL_STOP)");
	PRINTF("%sError stopping %s driver service!\n", szPrefix, szDriverName);
        // ignore any errors
    }

    if (DeleteService(hService)) {
	if (bVerbose) PRINTF("%s%s driver service deleted\n", szPrefix, szDriverName);
    } else {
	ShowErr("DeleteService");
	PRINTF("%sError while deleting %s driver service! (%s not removed)\n", szPrefix,
                  szDriverName, szDriverName);
	goto abort;
    }

    if (bDelDriverFile) {
        if (!MakeFullDriverFileName(szDriverName,
                            szDriverFileName, sizeof(szDriverFileName))) {
            ShowErr("MakeFullDriverFileName");
            PRINTF("error making driver file name!\n");
            goto finis;
        }
        // open driver file (just to see if it exists)
        if (!(hFile= CreateFile(szDriverFileName,
                         GENERIC_READ,
                         0,
                         (LPSECURITY_ATTRIBUTES)NULL,
                         CREATE_ALWAYS,
                         FILE_ATTRIBUTE_NORMAL,
                         NULL))) {
            goto finis;  // error...quietly exit
        } else {
            CloseHandle(hFile);
        }
        // delete the driver's .SYS file
        if (!DeleteFile(szDriverFileName)) {
            ShowErr("szDriverFileName");
            PRINTF("%sfailed to delete file %s!\n", szPrefix, szDriverFileName);
            goto abort;
        } else {
            if (bVerbose) PRINTF("%sdeleted driver file %s\n", szPrefix, szDriverFileName);
        }
    }

finis:
    if (bVerbose) PRINTF("%s%s driver service now may be re-installed\n", szPrefix, szDriverName);
    nExitStatus = TRUE;

  abort:
    if (hService) CloseServiceHandle(hService);
    if (hSCManager) CloseServiceHandle(hSCManager);
    return (nExitStatus);
}
