// SetPrivIoXP.c
//;-------------------------------------------------------------------------
//; Copyright (c) 2003 Steven Freyder, David Helland, and Bruce D. Lightner.
//; Use subject to licensing restrictions described in "PICOWEB SERVER
//; SOFTWARE LICENSE AGREEMENT DATED JULY 14, 1999" (See file LICENSE.tt).
//;-------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <winioctl.h>
#include <process.h>

#include "PrivIoXP_IOCTL.h"
#include "InstDriverXP.h"
#include "SetPrivIoXP.h"

extern unsigned char PPortIoDExe[];
extern int PPortIoDExeLen;

#ifdef WIN32
#define snprintf _snprintf
#endif

#define ENV_DAEMON_EXE "PPORTIOD_EXE"

#define PRINTF PrintfPrivIoXP

static void PathToExecutable(char *szFileName, int nFileNameLen)
{
    char *p = szFileName + strlen(szFileName);

    while ((p - szFileName) > 0) {  // back-scan for "/" or "\"
        if (*p == '\\' || *p == '/') break;
        --p;
    }
    if ((p - szFileName) > 0) ++p;  // restore "/" or "\"
    *p = '\0';  // truncate string here
    strncat(szFileName, PRIVEIOXP_PORT_DAEMON, nFileNameLen);
}

static BOOL RunDaemonPrivIoXP(
    BOOL bVerbose,      // 0: errors only, !=0: lots of diagnostics
    char *szPrefix,     // prefix string for error/diagnostic messages
    int nPort
){
    int nPID;
    char *p, *program;
    char szCommand[MAX_PATH];
    char szPort[16];
    HANDLE hFile = INVALID_HANDLE_VALUE;

    _snprintf(szPort, sizeof(szPort), "%d", nPort);

    // look to environment strings for executable
    program = getenv(ENV_DAEMON_EXE);
    if (program && *program) {
        nPID = _spawnlp(_P_DETACH, program, program, szPort, NULL);
        if (nPID >= 0) {
            if (bVerbose) PRINTF("%sspawned %s %s\n",
                                 szPrefix, program, szPort);
            return TRUE;
        } else {
            PRINTF("%sFailed to spawn $%s (%s)!\n",
                szPrefix, ENV_DAEMON_EXE, program);
            return FALSE;
        }
    }

    // look for executable in same place as our program
    if (GetModuleFileName(NULL, szCommand, sizeof(szCommand)) > 0) {
        PathToExecutable(szCommand, sizeof(szCommand));
        nPID = _spawnl(_P_DETACH, szCommand, szCommand, szPort, NULL);
        if (nPID >= 0) {
            if (bVerbose) PRINTF("%sspawned %s %s\n",
                                 szPrefix, szCommand, szPort);
            return TRUE;
        }
    }

    // look for executable in current path
    p = GetCommandLine();
    strncpy(szCommand, (p) ? p : "", sizeof(szCommand));
    PathToExecutable(szCommand, sizeof(szCommand));
    nPID = _spawnl(_P_DETACH, szCommand, szCommand, szPort, NULL);
    if (nPID >= 0) {
        if (bVerbose) PRINTF("%sspawned %s %s\n", szPrefix, szCommand, szPort);
        return TRUE;
    }

    // try default executable using PATH
    program = PRIVEIOXP_PORT_DAEMON;
    nPID = _spawnlp(_P_DETACH, program, program, szPort, NULL);
    if (nPID >= 0) {
        if (bVerbose) PRINTF("%sspawned %s %s\n", szPrefix, program, szPort);
        return TRUE;
    }

#ifndef DO_NOT_INSTALL_EXE_FILE
    // create executable and try once more
    if (PPortIoDExeLen &&
                GetModuleFileName(NULL, szCommand, sizeof(szCommand)) > 0) {
        PathToExecutable(szCommand, sizeof(szCommand));
        if (!(hFile = CreateFile(szCommand,
                         GENERIC_WRITE,
                         0,
                         (LPSECURITY_ATTRIBUTES)NULL,
                         CREATE_ALWAYS,
                         FILE_ATTRIBUTE_NORMAL,
                         NULL))) {
            ShowErr("szCommand");
            PRINTF("%sFailed to copy %s to %s!\n",
                   szPrefix, PRIVEIOXP_PORT_DAEMON, szCommand);
        } else {
            DWORD dwNumBytesWritten;

            if (!WriteFile(hFile, PPortIoDExe, PPortIoDExeLen,
                                                 &dwNumBytesWritten, 0)) {
                ShowErr("szCommand");
                PRINTF("%sFailed to copy %s to %s!\n",
                       szPrefix, PRIVEIOXP_PORT_DAEMON, szCommand);
            } else {
                if (!CloseHandle(hFile)) {
                    ShowErr("szCommand");
                    PRINTF("%sFailed to close file %s!\n", szPrefix, szCommand);
                } else {
                    PRINTF("%s%s installed at %s\n",
                           szPrefix, PRIVEIOXP_PORT_DAEMON, szCommand);
                    nPID = _spawnl(_P_DETACH, 
                                   szCommand, szCommand, szPort, NULL);
                    if (nPID >= 0) {
                        if (bVerbose) PRINTF("%sspawned %s %s\n",
                                             szPrefix, szCommand, szPort);
                        return TRUE;
                    }
                }
            }
        }
    }
#endif /* !DO_NOT_INSTALL_EXE_FILE */

    if (bVerbose) PRINTF("%sFailed to spawn %s %s\n",
                         szPrefix, program, szPort);
    return FALSE;
}

BOOL bIsWindowsNT(void)
{
    OSVERSIONINFO osvi;

    ZeroMemory(&osvi, sizeof(osvi));
    osvi.dwOSVersionInfoSize = sizeof(osvi);
    if (!GetVersionEx(&osvi)) {
        return FALSE;
    } 
    if (osvi.dwPlatformId == VER_PLATFORM_WIN32_NT) {
        return TRUE;
    } else {
        return FALSE;
    }
}

char *MakePortNamePrivIoXP(
    int nPort,
    char *szDeviceName,
    int nMaxBytes
){
    snprintf(szDeviceName, nMaxBytes,
                      "\\\\.\\%s%d", PRIVIOXP_DEVICE_BASE_NAME, nPort);
    return szDeviceName;
}

HANDLE OpenPortPrivIoXP(
    BOOL bVerbose,      // 0: errors only, !=0: lots of diagnostics
    char *szPrefix,     // prefix string for error/diagnostic messages
    int nPort,          // LPT port: 1=LPT1, 2=LPT2, etc.
    int *nOpenCount,    // NULL or returned device open count (1-n)
    int *nIoBase        // NULL or returned base I/O address of port
){
    HANDLE hDriver = NULL;
    char szDeviceName[MAX_PATH];
    int tries;
    long count = 0;
    long base = 0;
    DWORD nBytesRet;

    if (nOpenCount) *nOpenCount = count;
    if (nIoBase) *nIoBase = base;

    MakePortNamePrivIoXP(nPort, szDeviceName, sizeof(szDeviceName));
    ///PRINTF("opening %s...\n", szDeviceName); fflush(stdout);
    hDriver = CreateFile(szDeviceName,
                        //GENERIC_READ | GENERIC_WRITE,
                        //FILE_SHARE_READ | FILE_SHARE_WRITE,
                        GENERIC_READ,
                        FILE_SHARE_READ,
                        NULL,
                        OPEN_EXISTING,
                        0,
                        NULL);
    if (hDriver == INVALID_HANDLE_VALUE) {
        if (bVerbose) {
            ShowErr("CreateFile");
            PRINTF("%sfailed to open device %s\n", szPrefix, szDeviceName);
        }
        return INVALID_HANDLE_VALUE;
    }
    if (bVerbose) {
        PRINTF("%sopened device %s\n", szPrefix, szDeviceName);
    }

    if (!DeviceIoControl(hDriver, IOCTL_PRIVIOXP_READ_OPEN_COUNT,
                                  NULL, 0,
                                  &count, sizeof(count),
                                  &nBytesRet, NULL)) {
        if (bVerbose) {
            ShowErr("DeviceIoControl(IOCTL_PRIVIOXP_READ_OPEN_COUNT)");
            PRINTF("%serror reading open count\n", szPrefix);
        }
        CloseHandle(hDriver);
        return INVALID_HANDLE_VALUE;
    }
    if (nOpenCount) *nOpenCount = count;
    if (bVerbose) {
        PRINTF("%sopen count = %d\n", szPrefix, count);
    }

    if (!DeviceIoControl(hDriver, IOCTL_PRIVIOXP_READ_PORT_BASE,
                                  NULL, 0,
                                  &base, sizeof(base),
                                  &nBytesRet, NULL)) {
        if (bVerbose) {
            ShowErr("DeviceIoControl(IOCTL_PRIVIOXP_READ_PORT_BASE)");
            PRINTF("%serror reading I/O base address\n", szPrefix);
        }
        CloseHandle(hDriver);
        return INVALID_HANDLE_VALUE;
    }
    if (nIoBase) *nIoBase = base;
    if (bVerbose) {
        PRINTF("%sI/O base = 0x%03x\n", szPrefix, base);
    }

    return hDriver;
}

HANDLE OpenPrivIoXP(
    BOOL bVerbose,      // 0: errors only, !=0: lots of diagnostics
    char *szPrefix,     // prefix string for error/diagnostic messages
    BOOL bAutoInstall,  // TRUE: auto-install driver, FALSE: don't try
    int nPort,          // LPT port: 0=any, 1=LPT1, 2=LPT2
    int *nActualPort,   // NULL or returned LPT port (1, 2,...)
    int *nOpenCount,    // NULL or returned device open count (1-n)
    int *nIoBase        // NULL or returned base I/O address of port
){
    HANDLE hDriver = NULL;
    char szDeviceName[1024];
    int port, try, nPortStart, nPortEnd;

    if (nActualPort) *nActualPort = 0;
    if (nIoBase) *nIoBase = 0;

    // open driver
    for (try = 1; try <= 2 ; ++try) {
        if (nPort > 0) {
            nPortStart = nPortEnd = nPort;
        } else {
            nPortStart = 1;
            nPortEnd = 4;
        }
        for (port = nPortStart; port <= nPortEnd; ++port) {
            hDriver = OpenPortPrivIoXP(
                          (try == 1 && bAutoInstall) ? FALSE : TRUE,
                          szPrefix, port, nOpenCount, nIoBase);
            if (hDriver != INVALID_HANDLE_VALUE) {
                if (nActualPort) *nActualPort = port;
		return hDriver;
	    }
        }
        if (try == 1 && bAutoInstall) {
            // try to start driver for device
            if (!StartDriverXP(bVerbose, szPrefix, DRIVER_NAME)) {
                break;
            } else {
                continue;  // try again
            }
        }
    }
    if (!bAutoInstall) {
        ShowErr(DRIVER_NAME);
            PRINTF(
               "%scan't open "DRIVER_NAME" driver! (Hint: Is driver loaded?)\n",
               szPrefix);
    }
    return INVALID_HANDLE_VALUE;
}

static HANDLE hPrivIO = INVALID_HANDLE_VALUE;

BOOL UnSetPrivIoXP(void)
{

    if (hPrivIO != INVALID_HANDLE_VALUE) {
        CloseHandle(hPrivIO);
    }
    hPrivIO = INVALID_HANDLE_VALUE;
    return TRUE;
}

BOOL SetPrivIoXP(
    BOOL bVerbose,
    char *szPrefix,
    BOOL bAutoInstall,
    int nPort,          // LPT port: 0=any, 1=LPT1, 2=LPT2
    int *nActualPort,   // NULL or returned LPT port (1, 2,...)
    int *nOpenCount,    // NULL or returned device open count (1-n)
    int *nIoBase,       // NULL or returned base I/O address of port
    DWORD *ioList,     // NULL => grant direct access default port
    int ioListEntries  // number of DWORDs in ioList
){
    HANDLE hDriver = NULL;
    int tries;
    int nMyOpenCount;
    int ix;
    DWORD value;
    DWORD dwProcessId;
    DWORD nBytesRet;

    if (!bIsWindowsNT()) {
        PRINTF(
            "%sError! Only works with Windows NT (Windows NT/2000/XP)!\n", szPrefix);
        return FALSE;
    }

    if (nOpenCount) *nOpenCount = 0;
    if ((hDriver = OpenPrivIoXP(bVerbose, szPrefix, bAutoInstall,
             nPort, &nPort, &nMyOpenCount, nIoBase)) == INVALID_HANDLE_VALUE) {
        goto abort;
    }
    if (nOpenCount) *nOpenCount = nMyOpenCount;
    hPrivIO = hDriver;

    if (nMyOpenCount == 1) {
        if (!RunDaemonPrivIoXP(bVerbose, szPrefix, nPort)) {
            PRINTF("%sWarning! Unable to run daemon %s (errno=%d)!\n",
                   szPrefix, PRIVEIOXP_PORT_DAEMON, errno);
        }
    }

    // process arguments and send to driver
    if (ioList && ioListEntries > 0) {
        for (ix = 0; ix < ioListEntries; ++ix) {
            value = ioList[ix];
            if (!DeviceIoControl(hDriver,
                                    IOCTL_SET_IOPM,
                                    &value, sizeof(value),
                                    NULL, 0, &nBytesRet, NULL)) {
                ShowErr("DeviceIoControl(IOCTL_SET_IOPM)");
                PRINTF(
                    "%serror granting direct access to I/O address 0x%03x\n",
                    szPrefix, value);
                goto abort;
            } else {
                if (bVerbose) PRINTF(
                    "%sI/O address 0x%03x granted direct access\n",
                    szPrefix, value);
            }
        }
    } else {
#if 0
        if (!DeviceIoControl(hDriver,
                                IOCTL_IOPM_ALLOW_ALL_ACCESS,
                                NULL, 0,
                                NULL, 0, &nBytesRet, NULL)) {
            ShowErr("DeviceIoControl(IOCTL_IOPM_ALLOW_ALL_ACCESS)");
            PRINTF(
                "%serror granting direct access to all I/O addresses\n",
                 szPrefix);
            goto abort;
        } else {
            if (bVerbose) PRINTF(
                    "%sgranted direct access to all I/O ports\n", szPrefix);
        }
#else
        ; // we already have access to default parallel port I/O space
#endif
    }

#if 0
    // give us I/O rights
    dwProcessId = GetCurrentProcessId();
    if (!DeviceIoControl(hDriver,
                            IOCTL_ENABLE_IOPM_ON_PROCESSID,
                            &dwProcessId, sizeof(dwProcessId),
                            NULL, 0, &nBytesRet, NULL)) {

        ShowErr("DeviceIoControl(IOCTL_ENABLE_IOPM_ON_PROCESSID)");
        PRINTF( "%serror granting direct access to our process\n", szPrefix);
        goto abort;
    }
#endif

    return TRUE;

abort:
    if (hDriver) CloseHandle(hDriver);
    return FALSE;
}

BOOL MakeFullDriverFileName(
    char *szDriverName,
    char szDriverFileName[],
    int lenDriverFileName
){
    ZeroMemory(szDriverFileName, lenDriverFileName);
    if (!GetSystemDirectory(szDriverFileName,
                   lenDriverFileName - strlen(szDriverName) - 16)) {
        return FALSE;
    }
    strcat(szDriverFileName, "\\drivers\\");
    strcat(szDriverFileName, szDriverName);
    strcat(szDriverFileName, ".sys");
    return TRUE;
}

void ShowErr(char *why)
{
    char *lpMsgBuf = NULL;

    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
                  FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(),
                  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                  (LPTSTR)&lpMsgBuf, 0, NULL);
    if (lpMsgBuf && *lpMsgBuf) {
        PRINTF("%s: error %lu: %s", why, GetLastError(), lpMsgBuf);
        LocalFree(lpMsgBuf);
    } else {
        lpMsgBuf = "";
        if (GetLastError() == 10060) lpMsgBuf = "no response";
        if (GetLastError() == 317) lpMsgBuf = "no response";
        PRINTF("%s: error %lu: %s\n", why, GetLastError(), lpMsgBuf);
    }
    fflush(stderr);
}
